#include "ALSADevice.hpp"
#include "Board.hpp"
#include "Utils.hpp"
#include "V4L2Device.hpp"
#include "forward-v4l2-ioctl.h"

#include <climits>
#include <cmath>
#include <csignal>
#include <cstring>
#include <iomanip>
#include <iostream>
#include <list>
#include <poll.h>
#include <tuple>

/*!
 * \dir ./
 * This example generates A-V delay test pattern
 */

namespace {
static const int BUFFER_COUNT = 8;
static const int TEST_PERIOD_FIELDS = 10;
// ALSA chunk size in samples, should be less than frame (field) period
static const int AUDIO_PERIOD_SIZE = 400;
static const int AUDIO_CHANNELS = 2;
static const snd_pcm_uframes_t AUDIO_SAMPLE_RATE = 48000;
}

static bool shouldStop = false;
static uint8_t* videoTest;
snd_pcm_uframes_t audioBufSize;
int32_t* audioBufSilence;
int32_t* audioBufSin;

void stopSignal(int sig)
{
    if ((sig == SIGTERM) || (sig == SIGINT))
        shouldStop = true;
}

v4l2_format getFormat(V4L2Device& dev)
{
    v4l2_format fmt;
    fmt.type = dev.type();
    if (dev.ioctl(VIDIOC_G_FMT, &fmt)) {
        std::cerr << "Cannot get video format: " << dev.errorString() << std::endl;
        return v4l2_format();
    }
    return fmt;
}

v4l2_format setFormat(V4L2Device& dev, const v4l2_format& fmt)
{
    v4l2_format newFmt = fmt;
    if (dev.ioctl(VIDIOC_S_FMT, &newFmt)) {
        std::cerr << "Cannot set video format: " << dev.errorString() << std::endl;
        return fmt;
    }
    return newFmt;
}

//! Get required audio samples per video frame
snd_pcm_uframes_t getAudioPerVideoFrame(V4L2Device& dev, const v4l2_pix_format& fmt)
{
    v4l2_dv_timings tim;
    if (dev.ioctl(VIDIOC_G_DV_TIMINGS, &tim))
        return 0;

    uint64_t denom = tim.bt.pixelclock;
    uint64_t num
        = AUDIO_SAMPLE_RATE * V4L2_DV_BT_FRAME_WIDTH(&tim.bt) * V4L2_DV_BT_FRAME_HEIGHT(&tim.bt);

    // FIXME: No /1.001 compensation
    if (tim.bt.interlaced && (fmt.field == V4L2_FIELD_ALTERNATE))
        denom *= 2;

    return num / denom;
}

v4l2_format setupVideoFormat(V4L2Device& dev)
{
    // Get current frame format
    v4l2_format fmt = getFormat(dev);

    // Set field interleave - both fields are in one buffer
    fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;
    // Each field are in separate buffer
    // fmt.fmt.pix.field = V4L2_FIELD_ALTERNATE;
    // 8-bit UYVY
    fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_UYVY;
    // 8-bit YUYV
    // fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
    // 10-bit v210
    // fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_V210;
    // Raw - whole SDI frame, with HANC, VANC, 10-bit dense packed
    // fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_SL_RAW;

    // Set and adjust format
    fmt = setFormat(dev, fmt);

    std::cout << "Playback frame format is " << fmt << std::endl;

    return fmt;
}

void preroll(V4L2Device& dev, const std::vector<V4L2Device::Buffer>& buffers, v4l2_pix_format fmt)
{
    // Fill playback queue
    struct v4l2_buffer v4l2buf;
    memset(&v4l2buf, 0, sizeof(v4l2buf));
    v4l2buf.type = dev.type();
    v4l2buf.memory = V4L2_MEMORY_MMAP;
    v4l2buf.field = V4L2_FIELD_ANY;

    for (auto buf : buffers) {
        v4l2buf.index = buf.index;
        if (fmt.field == V4L2_FIELD_ALTERNATE)
            v4l2buf.field = (v4l2buf.index % 2) ? V4L2_FIELD_BOTTOM : V4L2_FIELD_TOP;

        dev.ioctl(VIDIOC_QBUF, &v4l2buf);
    }
}

void prerollAudio(ALSADevice& dev, int audioBufSize, int buffers)
{
    int32_t* audio_buf = new int32_t[audioBufSize * AUDIO_CHANNELS];
    memset(audio_buf, 0, audioBufSize * AUDIO_CHANNELS * sizeof(int32_t));

    for (int i = 0; i < buffers; i++)
        snd_pcm_writei(dev.pcm().get(), audio_buf, audioBufSize);

    delete[] audio_buf;
}

void setupAudio(ALSADevice& dev)
{
    snd_pcm_hw_params_t* params = nullptr;
    snd_pcm_sw_params_t* sw_params = nullptr;
    snd_pcm_format_t format;
    snd_pcm_uframes_t period = AUDIO_PERIOD_SIZE;
    unsigned int channels, rate;
    int dir;

    // Alloc hardware parameter structs
    snd_pcm_hw_params_malloc(&params);

    // Get default hardware parameters
    snd_pcm_hw_params_any(dev.pcm().get(), params);

    // Interleave channels
    snd_pcm_hw_params_set_access(dev.pcm().get(), params, SND_PCM_ACCESS_RW_INTERLEAVED);

    // Set period
    snd_pcm_hw_params_set_period_size_near(dev.pcm().get(), params, &period, &dir);

    // Set number of channels
    snd_pcm_hw_params_set_channels(dev.pcm().get(), params, AUDIO_CHANNELS);

    // Set hardware parameters, you can also change format, channels and other params prior this
    // call
    snd_pcm_hw_params(dev.pcm().get(), params);

    // Extract accepted parameters,
    snd_pcm_hw_params_get_format(params, &format);
    snd_pcm_hw_params_get_channels(params, &channels);
    snd_pcm_hw_params_get_rate(params, &rate, &dir);

    snd_pcm_hw_params_free(params);

    // Alloc software parameter structs
    snd_pcm_sw_params_alloca(&sw_params);
    // Get current software parameters
    snd_pcm_sw_params_current(dev.pcm().get(), sw_params);
    // Enable ALSA timestamping so we can always sync audio with video
    snd_pcm_sw_params_set_tstamp_mode(dev.pcm().get(), sw_params, SND_PCM_TSTAMP_ENABLE);
    // Disable ALSA auto-start
    snd_pcm_sw_params_set_start_threshold(dev.pcm().get(), sw_params, LONG_MAX);
    // Set software parameters
    snd_pcm_sw_params(dev.pcm().get(), sw_params);

    snd_pcm_prepare(dev.pcm().get());

    std::cout << "ALSA playback " << channels << " x " << snd_pcm_format_name(format) << " x "
              << rate << " with period of " << period << " samples" << std::endl;
}

void mainLoop(V4L2Device& videoDev, ALSADevice& audioDev,
    std::vector<V4L2Device::Buffer>& videoBuffers, const v4l2_pix_format& videoFormat)
{
    int videoType = videoDev.type();

    struct v4l2_buffer videoV4L2Buf;
    bool bottomField = false;
    int num = 0;

    memset(&videoV4L2Buf, 0, sizeof(videoV4L2Buf));
    videoV4L2Buf.type = videoType;
    videoV4L2Buf.memory = V4L2_MEMORY_MMAP;
    videoV4L2Buf.field = V4L2_FIELD_ANY;

    // Start capture, VBI device first!
    videoDev.ioctl(VIDIOC_STREAMON, &videoType);
    // Start audio capture
    snd_pcm_start(audioDev.pcm().get());

    pollfd pollfd[1];

    pollfd[0].fd = videoDev.fd();
    pollfd[0].events = POLLOUT;

    while (!shouldStop) {
        // Wait for frame
        poll(pollfd, 1, 100);

        timespec timestamp;
        clock_gettime(CLOCK_MONOTONIC, &timestamp);
        uint64_t curTs = (uint64_t)timestamp.tv_sec * 1000000000ULL + timestamp.tv_nsec;

        if (pollfd[0].revents & (POLLERR | POLLHUP | POLLNVAL))
            break;

        // Video/VBI frame sended
        if (pollfd[0].revents & POLLOUT) {
            // Get buffer from driver, video first.
            // Video and VBI is always in sync if number of queued buffers are always same
            // To be sure, you can check sequence fields
            videoDev.ioctl(VIDIOC_DQBUF, &videoV4L2Buf);

            uint64_t vidTs = (uint64_t)videoV4L2Buf.timestamp.tv_sec * 1000000000ULL
                + videoV4L2Buf.timestamp.tv_usec * 1000;

            std::cout << "Sended video buffer at " << curTs << "ns, "
                      << "video id = " << videoV4L2Buf.index
                      << ", video sequence = " << videoV4L2Buf.sequence
                      << ", video field = " << videoV4L2Buf.field << ", video timestamp = " << vidTs
                      << "ns, video size = " << videoV4L2Buf.bytesused << std::endl;

            uint8_t* videoData = videoBuffers[videoV4L2Buf.index].data;
            int32_t* audioData = nullptr;

            if (videoFormat.field == V4L2_FIELD_ALTERNATE) {
                videoV4L2Buf.field = bottomField ? V4L2_FIELD_BOTTOM : V4L2_FIELD_TOP;
                bottomField = !bottomField;

                int n = num % (TEST_PERIOD_FIELDS);
                uint8_t nY = (uint8_t)(16 + n * 235 / TEST_PERIOD_FIELDS);
                for (int i = 0; i < videoV4L2Buf.bytesused / 2; i++) {
                    videoData[i * 2 + 0] = 0x80;
                    videoData[i * 2 + 1] = nY;
                }
                audioData = ((num % TEST_PERIOD_FIELDS) == 0) ? audioBufSin : audioBufSilence;
            } else {
                int n = num % (TEST_PERIOD_FIELDS / 2) * 2;
                uint8_t nYO = (uint8_t)(16 + (n + 0) * 235 / TEST_PERIOD_FIELDS);
                uint8_t nYE = (uint8_t)(16 + (n + 1) * 235 / TEST_PERIOD_FIELDS);

                for (int l = 0; l < videoFormat.height; l++) {
                    for (int x = 0; x < videoFormat.width; x++) {
                        videoData[l * videoFormat.bytesperline + x * 2 + 0] = 0x80;
                        videoData[l * videoFormat.bytesperline + x * 2 + 1] = (l % 2) ? nYE : nYO;
                    }
                }
                audioData = ((num % (TEST_PERIOD_FIELDS / 2)) == 0) ? audioBufSin : audioBufSilence;
            }

            snd_pcm_sframes_t size = snd_pcm_writei(audioDev.pcm().get(), audioData, audioBufSize);

            std::cout << "Sending " << size << " audio frames at " << curTs << "ns, " << std::endl;

            videoDev.ioctl(VIDIOC_QBUF, &videoV4L2Buf);

            num++;
        }
    }

    // Stop audio capture
    snd_pcm_drop(audioDev.pcm().get());
    // Stop capture, Video device first!
    videoDev.ioctl(VIDIOC_STREAMOFF, &videoType);
}

int main(int argc, char** argv)
{
    std::string videoPath, audioPath;
    if (argc != 3) {
        std::cout << "Usage: " << argv[0] << " video_device audio_device" << std::endl;
        return 0;
    }

    videoPath = argv[1];
    std::cout << "Openning " << videoPath << std::endl;
    V4L2Device videoDev(videoPath);

    if (!videoDev.isOpen()) {
        std::cerr << "Cannot open video device: " << videoDev.errorString() << std::endl;
        return -1;
    }

    if (videoDev.type() != V4L2_BUF_TYPE_VIDEO_OUTPUT) {
        std::cerr << "Device " << videoPath << " is not an output video device" << std::endl;
        return -1;
    }

    audioPath = argv[2];
    ALSADevice audioDev(audioPath, SND_PCM_STREAM_PLAYBACK);

    if (!audioDev.isOpen()) {
        std::cerr << "Cannot open audio device: " << audioDev.errorString() << std::endl;
        return -1;
    }

    setupAudio(audioDev);

    signal(SIGTERM, &stopSignal);
    signal(SIGINT, &stopSignal);

    std::vector<V4L2Device::Buffer> videoBuffers;
    v4l2_format videoFormat;

    videoFormat = setupVideoFormat(videoDev);
    videoBuffers = videoDev.requestBuffers(BUFFER_COUNT);

    // Alloc audio buffer
    audioBufSize = getAudioPerVideoFrame(videoDev, videoFormat.fmt.pix);
    audioBufSilence = new int32_t[audioBufSize * AUDIO_CHANNELS];
    audioBufSin = new int32_t[audioBufSize * AUDIO_CHANNELS];

    memset(audioBufSin, 0, audioBufSize * AUDIO_CHANNELS * 4);
    memset(audioBufSilence, 0, audioBufSize * AUDIO_CHANNELS * 4);

    for (snd_pcm_uframes_t i = 0; i < 96; i++) {
        float v = (float)sin(M_PI * (i + 1) / 48);
        audioBufSin[i * 2 + 0] = v * (float)((1U << 31) - 1);
        audioBufSin[i * 2 + 1] = v * (float)((1U << 31) - 1);
    }

    std::cout << "Playback with " << videoBuffers.size() << " buffers" << std::endl;

    // Preroll video
    preroll(videoDev, videoBuffers, videoFormat.fmt.pix);
    prerollAudio(
        audioDev, getAudioPerVideoFrame(videoDev, videoFormat.fmt.pix), videoBuffers.size());

    mainLoop(videoDev, audioDev, videoBuffers, videoFormat.fmt.pix);

    delete[] audioBufSilence;
    delete[] audioBufSin;

    videoDev.freeBuffers(videoBuffers);

    return 0;
}
