#include "Board.hpp"
#include "Utils.hpp"
#include "V4L2Device.hpp"

#include <csignal>
#include <cstring>
#include <iomanip>
#include <iostream>
#include <poll.h>

/*!
 * \dir ./
 * This example shows simple video playback example
 */

namespace {
static const int BUFFER_COUNT = 4;
}

static bool shouldStop = false;
uint8_t* colorbar;

void stopSignal(int sig)
{
    if ((sig == SIGTERM) || (sig == SIGINT))
        shouldStop = true;
}

v4l2_format getFormat(V4L2Device& dev)
{
    v4l2_format fmt;
    fmt.type = dev.type();
    if (dev.ioctl(VIDIOC_G_FMT, &fmt)) {
        std::cerr << "Cannot get video format: " << dev.errorString() << std::endl;
        return v4l2_format();
    }
    return fmt;
}

v4l2_format setFormat(V4L2Device& dev, const v4l2_format& fmt)
{
    v4l2_format newFmt = fmt;
    if (dev.ioctl(VIDIOC_S_FMT, &newFmt)) {
        std::cerr << "Cannot set video format: " << dev.errorString() << std::endl;
        return fmt;
    }
    return newFmt;
}

v4l2_format setupFormat(V4L2Device& dev)
{
    // Get current frame format
    v4l2_format fmt = getFormat(dev);

    // Set field interleave - both fields are in one buffer
    fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;
    // Each field are in separate buffer
    // fmt.fmt.pix.field = V4L2_FIELD_ALTERNATE;

    // 8-bit UYVY
    fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_UYVY;
    // 8-bit YUYV
    // fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
    // 10-bit v210
    // fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_V210;
    // Raw - whole SDI frame, with HANC, VANC, 10-bit dense packed
    // fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_SL_RAW;

    // Set and adjust format
    fmt = setFormat(dev, fmt);

    std::cout << "Playback frame format is " << fmt << std::endl;

    return fmt;
}

void preroll(
    V4L2Device& dev, const std::vector<V4L2Device::Buffer>& buffers, const v4l2_pix_format& fmt)
{
    // Fill playback queue
    struct v4l2_buffer v4l2buf;
    memset(&v4l2buf, 0, sizeof(v4l2buf));
    v4l2buf.type = dev.type();
    v4l2buf.memory = V4L2_MEMORY_MMAP;
    v4l2buf.field = V4L2_FIELD_ANY;

    for (auto buf : buffers) {
        v4l2buf.index = buf.index;
        if (fmt.field == V4L2_FIELD_ALTERNATE)
            v4l2buf.field = (v4l2buf.index % 2) ? V4L2_FIELD_BOTTOM : V4L2_FIELD_TOP;

        memcpy(buffers[v4l2buf.index].data, colorbar, fmt.sizeimage);

        dev.ioctl(VIDIOC_QBUF, &v4l2buf);
    }
}

void mainLoop(V4L2Device& dev, std::vector<V4L2Device::Buffer>& buffers, const v4l2_pix_format& fmt)
{
    int type = dev.type();

    struct v4l2_buffer v4l2buf;
    memset(&v4l2buf, 0, sizeof(v4l2buf));
    v4l2buf.type = type;
    v4l2buf.memory = V4L2_MEMORY_MMAP;
    v4l2buf.field = V4L2_FIELD_ANY;
    bool bottomField = false;

    // Start capture
    dev.ioctl(VIDIOC_STREAMON, &type);

    pollfd pollfd;
    pollfd.fd = dev.fd();
    pollfd.events = POLLOUT;

    while (!shouldStop) {
        // Wait for frame
        poll(&pollfd, 1, 100);

        if (pollfd.revents & (POLLERR | POLLHUP | POLLNVAL))
            break;

        // Timeout, no frame
        if (!(pollfd.revents & POLLOUT))
            continue;

        // Get buffer from queue
        dev.ioctl(VIDIOC_DQBUF, &v4l2buf);

        /*
         * PROCESS BUFFER DATA HERE
         */

        memcpy(buffers[v4l2buf.index].data, colorbar, fmt.sizeimage);
        // Don't forget to set correct field for V4L2_FIELD_ALTERNATE mode
        if (fmt.field == V4L2_FIELD_ALTERNATE) {
            v4l2buf.field = bottomField ? V4L2_FIELD_BOTTOM : V4L2_FIELD_TOP;
            bottomField = !bottomField;
        }

        timespec timestamp;
        clock_gettime(CLOCK_MONOTONIC, &timestamp);

        std::cout << "Sending buffer at "
                  << (uint64_t)timestamp.tv_sec * 1000000000ULL + timestamp.tv_nsec
                  << "ns, id = " << v4l2buf.index << ", sequence = " << v4l2buf.sequence
                  << ", field = " << v4l2buf.field << ", timestamp = "
                  << (uint64_t)v4l2buf.timestamp.tv_sec * 1000000000ULL
                + v4l2buf.timestamp.tv_usec * 1000
                  << "ns, size = " << v4l2buf.bytesused << std::endl;

        // Return buffer back to queue
        dev.ioctl(VIDIOC_QBUF, &v4l2buf);
    }

    // Stop capture
    dev.ioctl(VIDIOC_STREAMOFF, &type);
}

int main(int argc, char** argv)
{
    std::string devPath;
    if (argc != 2) {
        std::cout << "Usage: " << argv[0] << " device" << std::endl;
        return 0;
    }

    devPath = argv[1];
    std::cout << "Openning " << devPath << std::endl;
    V4L2Device dev(devPath);

    if (!dev.isOpen()) {
        std::cerr << "Cannot open device: " << dev.errorString() << std::endl;
        return -1;
    }

    if (dev.type() != V4L2_BUF_TYPE_VIDEO_OUTPUT) {
        std::cerr << "Device (" << dev.name() << ") is not an output device" << std::endl;
        return -1;
    }

    signal(SIGTERM, &stopSignal);
    signal(SIGINT, &stopSignal);

    std::vector<V4L2Device::Buffer> buffers;
    v4l2_format format;

    format = setupFormat(dev);
    buffers = dev.requestBuffers(BUFFER_COUNT);

    colorbar = new uint8_t[format.fmt.pix.sizeimage];
    generateRP219ColorBarUYVY(colorbar, format.fmt.pix.width, format.fmt.pix.height);

    std::cout << "Playback with " << buffers.size() << " buffers" << std::endl;

    preroll(dev, buffers, format.fmt.pix);
    mainLoop(dev, buffers, format.fmt.pix);

    dev.freeBuffers(buffers);

    return 0;
}
